#include <time.h>
#include <ctype.h>
#include <stdio.h>
#if defined(sgi)
#include <alloca.h>
#endif
#include <string.h>
#include <stdlib.h>

extern char* tzname[2];
#define HAVE_MKTIME

#ifndef Bool
#define Bool int
#endif

#ifndef True
#define True 1
#endif

#ifndef False
#define False 0
#endif

#define TIMEZONEERR 99999999

struct timezone_table_t
{
    char* name;
    long timezone;
};


char* local_zone_name = NULL;
static long get_local_timezone(void);

struct timezone_table_t timezone_table[] =
{
    {"PST", -8}, {"PDT", -7}, {"MST", -7}, {"MDT", -6}, {"CST", -6},
    {"CDT", -5}, {"EST", -5}, {"EDT", -4}, {"GMT", 0}, {"BST", 1},
    {"MET", 1}, {"EET", 2}, {"DST", 2}, {"JST", 9}, {"GMT+1", 1}, {"GMT+2", 2},
    {"GMT+3", 3}, {"GMT+4", 4}, {"GMT+5", 5}, {"GMT+6", 6}, {"GMT+7", 7},
    {"GMT+8", 8}, {"GMT+9", 9}, {"GMT+10", 10}, {"GMT+11", 11}, {"GMT+12", 12},
    {"GMT+13", 13}, {"GMT-1", -1}, {"GMT-2", -2}, {"GMT-3", -3},
    {"GMT-4",-4}, {"GMT-5", -5}, {"GMT-6", -6}, {"GMT-7", -7}, {"GMT-8", -8},
    {"GMT-9", -9}, {"GMT-10", -10}, {"GMT-11", -11}, {"GMT-12", -12},
    {NULL, 0}};

char* month_names[] =
{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", NULL};


/* This is copy from wcol(version 0.0722) in http.c, wcol.h,
 * and modified by Masanao Izumo<mo@goice.co.jp>
 */
#define strcasecmp_first_all(x, y0, y) \
			((x[0]==y0 || tolower(x[0])==tolower(y0)) && \
			strcasecmp(x,y)==0)

static int parse_month(const char *m)
{
    int i;

    if(m == NULL)
	return -1;

    for(i = 0; i < 12; i++)
	if(strcasecmp(m, month_names[i]) == 0)
	    return i;
    return -1;
}

long timezone_string_to_long(const char* name)
{
    int i;

    if(name == NULL)
	return get_local_timezone();

  retry_parse:
    if(name[0] == '+' || name[0] == '-')
    {
	long h, m;

#define CTOD(s) ((s) - '0')
	h = CTOD(name[1]) * 10 + CTOD(name[2]);
	m = CTOD(name[3]) * 10 + CTOD(name[4]);
#undef CTOD

	if(name[0] == '+')
	    return 3600 * h + 60 * m;
	return -3600 * h + 60 * m;
    }


    for(i = 0; timezone_table[i].name != NULL; i++)
	if(strcmp(timezone_table[i].name, name) == 0)
	    return timezone_table[i].timezone * 3600;
    if(name[0] == 'G' && name[1] == 'M' && name[2] == 'T' &&
       (name[3] == '+' || name[3] == '-'))
    {
	name += 3;
	goto retry_parse;
    }

    return TIMEZONEERR;
}

static Bool parse_hms(const char* hmc, int* h, int* m, int* c)
{
    if(hmc == NULL)
	return False;
    return sscanf(hmc, "%d:%d:%d", h, m, c) == 3;
}

static Bool parse_dmy(const char* dmy, int* day, int* month, int* year)
{
    char month_string[4];
    if(dmy == NULL)
	return False;
    if(sscanf(dmy, "%d-%c%c%c-%d",
	      day,
	      &month_string[0], &month_string[1], &month_string[2],
	      year) != 5)
	return False;
    month_string[3] = '\0';
    if((*month = parse_month(month_string)) < 0)
    {
	*month = 0;
	return False;
    }
    return True;
}

static int atoisafe(const char* s)
{
    if(s == NULL)
	return 0;
    return atoi(s);
}

int split_args(char* argbuff, int maxargc, char** argv)
{
    char* argstart;
    char* argend;
    int argc;

    argstart = argbuff;
    maxargc--;

    for(argc = 0; argc < maxargc; argc++)
    {
	/* skip whitespace */
	while(*argstart != '\0' && (*argstart == ' ' || *argstart == '\t'))
	    argstart++;

	if(*argstart == '\0')
	    break;
	argv[argc] = argstart;

	/* find end of arg. */
	argend = argstart;
	while(*argend != '\0' && *argend != ' ' && *argend != '\t')
	    argend++;
	if(*argend == '\0')
	{
	    argc++;
	    break;
	}
	*argend = '\0';
	argstart = argend + 1;
    }
    argv[argc] = NULL;
    return argc;
}

#define NORMALIZE_YEAR(y) ((y) -= ((y) >= 1900 ? 1900 : 0))
#define DATE_MAX_ARG 10
Bool parse_date_string(const char *date_arg,
		       int* sec, int* min, int* hour,
		       int* day, int* mon, int* year, long* timezone)
{
    char* date;
    size_t n;
    int i;
    char *date_argv[DATE_MAX_ARG];

    /*
     * asctime_date:
     *	week	month	day		hour:min:sec timezone year
     *	week	month	day		hour:min:sec year
     *	week	month	day		hour:min:sec year timezone ?
     *
     * rfc850_date:
     *	week	d-m-y	hour:min:sec	[timezone]
     *
     * rfc1123_date1:
     *	week,	day	month		year hour:min:sec [timezone]
     *	week	day	month		year hour:min:sec [timezone]
     *
     * rfc1123_date2:
     *	day	month	year		hour:min:sec [timezone]
     *
     * japanese date: (Not suppoted)
     * yyyyǯmmdd () hhmmʬss timezone
     */
    n = strlen(date_arg);
    date = (char *)alloca(n + 1);
    memcpy(date, date_arg, n + 1);
    for(i = 0; i < n; i++)
	if(date[i] == ',')
	    date[i] = ' ';
    while(n > 0 && !isgraph(date[n - 1]))
	date[--n] = '\0';

    if(split_args(date, DATE_MAX_ARG, date_argv) < 3)
	return False;

    if(strchr(date_argv[1], '-'))
	goto rfc850_date;

    if(isdigit(date_argv[0][0]))
	goto rfc1123_date2;

    if(isdigit(date_argv[2][0]))
	goto asctime_date;
    
    goto rfc1123_date1;

  asctime_date:
    /*	week	month	day	hour:min:sec year [timezone] */
    *mon = parse_month(date_argv[1]);
    *day   = atoisafe(date_argv[2]);
    if(!parse_hms(date_argv[3], hour, min, sec))
	return False;
    if(isdigit(date_argv[4][0]))
    {
	*year  = atoisafe(date_argv[4]);
	NORMALIZE_YEAR(*year);
	if((*timezone = timezone_string_to_long(date_argv[5])) == TIMEZONEERR)
	    return False;
	return True;
    }
    if((*timezone = timezone_string_to_long(date_argv[4])) == TIMEZONEERR)
	return False;
    *year  = atoisafe(date_argv[5]);
    NORMALIZE_YEAR(*year);
    return True;

  rfc850_date:
    /*	week	d-m-y	hour:min:sec [timezone] */
    if(!parse_dmy(date_argv[1], day, mon, year))
	return False;
    if(!parse_hms(date_argv[2], hour, min, sec))
	return False;
    if((*timezone = timezone_string_to_long(date_argv[3])) == TIMEZONEERR)
	return False;
    return True;

  rfc1123_date1:
    /*	week,	day	month	year hour:min:sec [timezone]
     *	week	day	month	year hour:min:sec [timezone] */
    *day = atoisafe(date_argv[1]);
    *mon = parse_month(date_argv[2]);
    *year = atoisafe(date_argv[3]);
    NORMALIZE_YEAR(*year);
    if(!parse_hms(date_argv[4], hour, min, sec))
	return False;
    if((*timezone = timezone_string_to_long(date_argv[5])) == TIMEZONEERR)
	return False;
    return True;

  rfc1123_date2:
    /*	day	month	year	hour:min:sec [timezone] */
    *day = atoisafe(date_argv[0]);
    *mon = parse_month(date_argv[1]);
    *year = atoisafe(date_argv[2]);
    NORMALIZE_YEAR(*year);
    if(!parse_hms(date_argv[3], hour, min, sec))
	return False;
    if((*timezone = timezone_string_to_long(date_argv[4])) == TIMEZONEERR)
	return False;
    return True;
}

static long get_local_timezone(void)
{
    if(local_zone_name != NULL)
	return timezone_string_to_long(local_zone_name);
    if(*tzname != NULL)
	return timezone_string_to_long(*tzname);
    tzset();
    if(*tzname == NULL)
	return 0;
    return timezone_string_to_long(*tzname);
}

time_t date_to_time(const char* date)
{
    struct tm TM;
    time_t ret;
    int year, mon, day;
    int hour, min, sec;
    long timezone;

    if(!parse_date_string(date, &sec, &min, &hour, &day, &mon, &year,
			  &timezone))
	return -1;
#if !defined(sgi)
    TM.tm_zone = NULL;
#endif
    TM.tm_year = year;
    TM.tm_mon  = mon;
    TM.tm_mday = day;
    TM.tm_hour = hour;
    TM.tm_min  = min;
    TM.tm_sec  = sec;
    TM.tm_isdst= 0;

#ifdef HAVE_MKTIME
    ret = mktime(&TM);
#else
    /*
     * for unbelievable system which not have mktime().
     */
    /* very very dirty (for save last one bit) */
    ret = sec/2 + min*30 + hour*30*60
            + day*30*60*24 + mon*30*60*24*31 + year*30*60*24*31*12;
#endif

    if(timezone == TIMEZONEERR)
	return ret;

    return ret - timezone + get_local_timezone();
}

int date_cmp(const char* date1, const char* date2)
{
    time_t t1, t2;

    t1 = date_to_time(date1);
    t2 = date_to_time(date2);

    if(t1 < t2)
	return -1;
    if(t1 > t2)
	return 1;
    return 0;
}

char* truncate_newline(char* s)
{
    int len = strlen(s);

    if(len < 2)
    {
	if(len == 0)
	    return s;
	if(s[0] == '\n' || s[0] == '\r')
	    s[0] = '\0';
	return s;
    }
    if(s[len - 1] == '\n')
	s[--len] = '\0';
    if(s[len - 1] == '\r')
	s[--len] = '\0';
    return s;
}


int main(int argc, char** argv)
{
    time_t t;
    char* date;

    if(argc != 2)
    {
	fprintf(stderr, "Usage: %s \"date\"\n", argv[0]);
	exit(1);
    }

    date = argv[1];

    truncate_newline(date);
    t = date_to_time(date);
    if(t == -1)
	fprintf(stderr, "Unrecognized: %s\n", date);
    printf("%lu\n", t);
    return 0;
}
